#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Jose Javier Merchante <jjmerchante@bitergia.com>
#

import datetime
import json
import os
import unittest.mock

import httpretty
from dateutil.tz import tzutc

from perceval.backend import BackendCommandArgumentParser
from perceval.backends.topicbox.topicbox import (Topicbox,
                                                 TopicboxCommand,
                                                 TopicboxClient)

TOPICBOX_GROUPS = 'data/topicbox/topicbox_groups.json'
TOPICBOX_MESSAGES = 'data/topicbox/topicbox_messages.json'
TOPICBOX_PAGE_1 = 'data/topicbox/topicbox_messages_page_1.json'
TOPICBOX_PAGE_2 = 'data/topicbox/topicbox_messages_page_2.json'

TOPICBOX_JMAP_URL = 'http://example.com/jmap'
TOPICBOX_GROUP_URL = 'http://example.com/groups/test_group'


def read_file(filename, mode='r'):
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), filename), mode) as f:
        content = f.read()
    return content


def setup_http_server():
    topicbox_msgs = read_file(TOPICBOX_MESSAGES)

    http_requests = []

    def request_callback(method, uri, headers):
        last_request = httpretty.last_request()
        body = topicbox_msgs
        http_requests.append(last_request)
        return 200, headers, body

    httpretty.register_uri(httpretty.POST,
                           TOPICBOX_JMAP_URL,
                           responses=[
                               httpretty.Response(body=request_callback)
                           ])

    return http_requests


class TestTopicboxBackend(unittest.TestCase):
    """Topicbox backend tests"""

    def test_initialization(self):
        """Test whether attributes are initialized"""

        topicbox = Topicbox(uri=TOPICBOX_GROUP_URL,
                            account_id='acc_id',
                            tag='test',
                            max_items=10)

        self.assertEqual(topicbox.uri, TOPICBOX_GROUP_URL)
        self.assertEqual(topicbox.tag, 'test')
        self.assertEqual(topicbox.account_id, 'acc_id')
        self.assertEqual(topicbox.group_tag, 'test_group')
        self.assertEqual(topicbox.max_items, 10)

        # When tag is empty or None it will be set to
        # the value in the origin
        topicbox = Topicbox(TOPICBOX_GROUP_URL, 'acc_id')
        self.assertEqual(topicbox.uri, TOPICBOX_GROUP_URL)
        self.assertEqual(topicbox.tag, TOPICBOX_GROUP_URL)

        Topicbox(TOPICBOX_GROUP_URL, 'acc_id', tag='')
        self.assertEqual(topicbox.uri, TOPICBOX_GROUP_URL)
        self.assertEqual(topicbox.tag, TOPICBOX_GROUP_URL)

    def test_has_archiving(self):
        """Test if it returns False when has_archiving is called"""

        self.assertEqual(Topicbox.has_archiving(), False)

    def test_has_resuming(self):
        """Test if it returns True when has_resuming is called"""

        self.assertEqual(Topicbox.has_resuming(), True)

    @httpretty.activate
    @unittest.mock.patch('perceval.backends.topicbox.topicbox.TopicboxClient.group')
    def test_fetch(self, group_patch):
        """Test whether it fetches a set of messages"""

        group_patch.return_value = {'archiveMailboxId': 'test_group'}
        setup_http_server()

        backend = Topicbox(TOPICBOX_GROUP_URL, 'acc_id')
        messages = [m for m in backend.fetch()]

        expected = [
            ['message-1', '1af8c390de5940fbbef8daab31dd2e23ae9a514f', 1619071380.0],
            ['message-2', 'd64fbfa6c18670f7b25d5cb0b9fc7db461c9fcee', 1619080429.0],
            ['message-3', '191bcf17f547d9d953c834895d0f3d6dfee42aac', 1619080613.0]
        ]

        self.assertEqual(len(messages), len(expected))

        for i, message in enumerate(messages):
            self.assertEqual(message['data']['id'], expected[i][0])
            self.assertEqual(message['data']['threadId'], 'Thread-id-1')
            self.assertEqual(message['origin'], TOPICBOX_GROUP_URL)
            self.assertEqual(message['uuid'], expected[i][1])
            self.assertEqual(message['updated_on'], expected[i][2])
            self.assertEqual(message['category'], 'message')
            self.assertEqual(message['tag'], TOPICBOX_GROUP_URL)

    @httpretty.activate
    @unittest.mock.patch('perceval.backends.topicbox.topicbox.TopicboxClient.group')
    def test_search_fields(self, group_patch):
        """Test whether the search_fields is properly set"""

        group_patch.return_value = {'archiveMailboxId': 'test_group'}
        setup_http_server()

        backend = Topicbox(TOPICBOX_GROUP_URL, 'acc_id')
        messages = [m for m in backend.fetch()]

        for message in messages:
            self.assertEqual(backend.metadata_id(message['data']), message['search_fields']['item_id'])


class TestTopicboxCommand(unittest.TestCase):
    """Tests for TopicboxCommand class"""

    def test_backend_class(self):
        """Test if the backend class is Topicbox"""

        self.assertIs(TopicboxCommand.BACKEND, Topicbox)

    def test_setup_cmd_parser(self):
        """Test if the parser object is correctly initialized"""

        parser = TopicboxCommand.setup_cmd_parser()
        self.assertIsInstance(parser, BackendCommandArgumentParser)
        self.assertEqual(parser._backend, Topicbox)

        args = [
            TOPICBOX_GROUP_URL,
            '--account-id', 'acc_id',
            '--tag', 'test',
            '--from-date', '2020-01-01',
            '--max-items', '10'
        ]
        parsed_args = parser.parse(*args)
        self.assertEqual(parsed_args.uri, TOPICBOX_GROUP_URL)
        self.assertEqual(parsed_args.account_id, 'acc_id')
        self.assertEqual(parsed_args.tag, 'test')
        self.assertEqual(parsed_args.max_items, 10)
        self.assertEqual(parsed_args.from_date, datetime.datetime(2020, 1, 1, tzinfo=tzutc()))


class TestTopicboxClient(unittest.TestCase):
    """Tests for Topicbox client class"""

    def test_init(self):
        """Test initialization"""

        client = TopicboxClient(base_uri='http://example.com',
                                account_id='account1',
                                max_items=30)

        self.assertIsInstance(client, TopicboxClient)
        self.assertEqual(client.account_id, 'account1')
        self.assertEqual(client.max_items, 30)

    @httpretty.activate
    def test_group(self):
        """Test find group id from identifier"""

        # Set up a mock HTTP server
        body = read_file(TOPICBOX_GROUPS)
        httpretty.register_uri(httpretty.POST,
                               TOPICBOX_JMAP_URL,
                               body=body, status=200)

        expected_1 = json.loads(body)['methodResponses'][0][1]['list'][0]
        expected_2 = json.loads(body)['methodResponses'][0][1]['list'][1]

        # Call API
        client = TopicboxClient(base_uri='http://example.com',
                                account_id='account1',
                                max_items=35)

        group_1 = client.group('mailbox1')
        self.assertEqual(group_1, expected_1)

        group_2 = client.group('mailbox2')
        self.assertEqual(group_2, expected_2)

    @httpretty.activate
    def test_emails(self):
        """Test Emails/query request"""

        # Set up a mock HTTP server
        body = read_file(TOPICBOX_MESSAGES)
        httpretty.register_uri(httpretty.POST,
                               TOPICBOX_JMAP_URL,
                               body=body, status=200)

        # Call API
        client = TopicboxClient(base_uri='http://example.com',
                                account_id='account1',
                                max_items=35)
        messages = client.messages('mailbox1')

        self.assertEqual(len(list(messages)), 3)

        # Check request payload
        request_payload = {
            "using": [
                "urn:ietf:params:jmap:mail",
                "urn:ietf:params:jmap:core",
                "https://www.topicbox.com/dev/organisation"
            ],
            "methodCalls": [
                [
                    "Email/query",
                    {
                        "accountId": "account1",
                        "calculateTotal": True,
                        "limit": 35,
                        "position": 0,
                        "collapseThreads": False,
                        "sort": [
                            {
                                "property": "receivedAt",
                                "isAscending": True
                            }
                        ],
                        "filter": {
                            "operator": "AND",
                            "conditions": [
                                {
                                    "inMailbox": "mailbox1"
                                },
                                {
                                    "after": "1970-01-01T00:00:00Z"
                                },
                                {
                                    "before": "2100-01-01T00:00:00Z"
                                }
                            ]
                        }
                    },
                    "0"
                ],
                [
                    "Email/get",
                    {
                        "accountId": "account1",
                        "#ids": {
                            "name": "Email/query",
                            "path": "/ids",
                            "resultOf": "0"
                        },
                        "fetchHTMLBodyValues": True,
                        "properties": None
                    },
                    "1"
                ]
            ]
        }

        req = httpretty.last_request()

        self.assertEqual(req.method, 'POST')
        self.assertRegex(req.path, '/jmap')
        self.assertDictEqual(json.loads(req.parsed_body), request_payload)

    @httpretty.activate
    def test_messages_pagination(self):
        """Test Emails/query request with pagination"""

        # Set up a mock HTTP server
        bodies = [
            read_file(TOPICBOX_PAGE_1),
            read_file(TOPICBOX_PAGE_2)
        ]

        http_requests = []

        def request_callback(method, uri, headers):
            last_request = httpretty.last_request()
            http_requests.append(last_request)
            body = bodies.pop(0)
            return [200, headers, body]

        httpretty.register_uri(httpretty.POST,
                               TOPICBOX_JMAP_URL,
                               body=request_callback,
                               status=200)

        # Call API
        client = TopicboxClient(base_uri='http://example.com',
                                account_id='account1',
                                max_items=2)
        messages = [m for m in client.messages('mailbox1')]

        self.assertEqual(len(list(messages)), 3)

        expected_messages = ['message-1', 'message-2', 'message-3']

        request_payload_1 = {
            "using": [
                "urn:ietf:params:jmap:mail",
                "urn:ietf:params:jmap:core",
                "https://www.topicbox.com/dev/organisation"
            ],
            "methodCalls": [
                [
                    "Email/query",
                    {
                        "accountId": "account1",
                        "calculateTotal": True,
                        "limit": 2,
                        "position": 0,
                        "collapseThreads": False,
                        "sort": [
                            {
                                "property": "receivedAt",
                                "isAscending": True
                            }
                        ],
                        "filter": {
                            "operator": "AND",
                            "conditions": [
                                {
                                    "inMailbox": "mailbox1"
                                },
                                {
                                    "after": "1970-01-01T00:00:00Z"
                                },
                                {
                                    "before": "2100-01-01T00:00:00Z"
                                }
                            ]
                        }
                    },
                    "0"
                ],
                [
                    "Email/get",
                    {
                        "accountId": "account1",
                        "#ids": {
                            "name": "Email/query",
                            "path": "/ids",
                            "resultOf": "0"
                        },
                        "fetchHTMLBodyValues": True,
                        "properties": None
                    },
                    "1"
                ]
            ]
        }

        request_payload_2 = {
            "using": [
                "urn:ietf:params:jmap:mail",
                "urn:ietf:params:jmap:core",
                "https://www.topicbox.com/dev/organisation"
            ],
            "methodCalls": [
                [
                    "Email/query",
                    {
                        "accountId": "account1",
                        "calculateTotal": True,
                        "limit": 2,
                        "position": 2,
                        "collapseThreads": False,
                        "sort": [
                            {
                                "property": "receivedAt",
                                "isAscending": True
                            }
                        ],
                        "filter": {
                            "operator": "AND",
                            "conditions": [
                                {
                                    "inMailbox": "mailbox1"
                                },
                                {
                                    "after": "1970-01-01T00:00:00Z"
                                },
                                {
                                    "before": "2100-01-01T00:00:00Z"
                                }
                            ]
                        }
                    },
                    "0"
                ],
                [
                    "Email/get",
                    {
                        "accountId": "account1",
                        "#ids": {
                            "name": "Email/query",
                            "path": "/ids",
                            "resultOf": "0"
                        },
                        "fetchHTMLBodyValues": True,
                        "properties": None
                    },
                    "1"
                ]
            ]
        }

        self.assertEqual(len(http_requests), 2)

        req = http_requests[0]
        self.assertEqual(req.method, 'POST')
        self.assertRegex(req.path, '/jmap')
        self.assertDictEqual(json.loads(req.parsed_body), request_payload_1)

        req = http_requests[1]
        self.assertEqual(req.method, 'POST')
        self.assertRegex(req.path, '/jmap')
        self.assertDictEqual(json.loads(req.parsed_body), request_payload_2)

        for i, message in enumerate(messages):
            self.assertEqual(message['id'], expected_messages[i])
            self.assertEqual(message['threadId'], 'Thread-id-1')


if __name__ == "__main__":
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Jose Javier Merchante <jjmerchante@bitergia.com>
#

import datetime
import json
import logging

import dateutil.tz

from grimoirelab_toolkit.datetime import str_to_datetime, datetime_to_utc
from grimoirelab_toolkit.uris import urijoin

from ...backend import (Backend,
                        BackendCommand,
                        BackendCommandArgumentParser)
from ...client import HttpClient
from ...errors import BackendError

CATEGORY_MESSAGE = 'message'

MAX_MESSAGES_PER_PAGE = 100

JMAP_CLIENT_CAPABILITIES = [
    "urn:ietf:params:jmap:mail",
    "urn:ietf:params:jmap:core",
    "https://www.topicbox.com/dev/organisation"
]

DEFAULT_DATETIME = datetime.datetime(1970, 1, 1, 0, 0, 0,
                                     tzinfo=dateutil.tz.tzutc())
DEFAULT_LAST_DATETIME = datetime.datetime(2100, 1, 1, 0, 0, 0,
                                          tzinfo=dateutil.tz.tzutc())

logger = logging.getLogger(__name__)


class Topicbox(Backend):
    """Topicbox backend for Perceval.

    This class retrieves the email messages stored in Topicbox group
    using the JSON Meta Application Protocol (JMAP).
    Initialize this class passing the URL of a Topicbox group.
    The origin of the data will be set to the value of `uri`.

    :param uri: URI of the Topicbox group
    :param account_id: account id for Topicbox's JMAP
    :param tag: label used to mark the data
    :param archive: archive to store/retrieve items
    """
    version = '0.1.0'

    CATEGORIES = [CATEGORY_MESSAGE]

    DATE_FIELD = 'sentAt'
    MESSAGE_ID_FIELD = 'messageId'

    def __init__(self, uri, account_id, tag=None, archive=None,
                 max_items=MAX_MESSAGES_PER_PAGE, ssl_verify=True):
        origin = uri

        super().__init__(origin, tag=tag, archive=archive, ssl_verify=ssl_verify)

        self.uri = uri
        self.group_tag = uri.rstrip('/').split('/')[-1]
        self.account_id = account_id
        self.max_items = max_items

        self.client = None

    def fetch(self, category=CATEGORY_MESSAGE, filter_classified=False,
              from_date=DEFAULT_DATETIME, to_date=DEFAULT_LAST_DATETIME):
        """Fetch the emails from a Topicbox group.

        The method retrieves from Topicbox the messages using the
        JMAP protocol after the given date.

        :param category: the category of items to fetch
        :param filter_classified: remove classified fields from the resulting items
        :param from_date: obtain messages created since this date
        :param to_date: obtain messages until a specific date

        :returns: a generator of messages
        """
        if not from_date:
            from_date = DEFAULT_DATETIME
        if not to_date:
            to_date = DEFAULT_LAST_DATETIME

        from_date = datetime_to_utc(from_date)
        to_date = datetime_to_utc(to_date)

        kwargs = {
            'from_date': from_date,
            'to_date': to_date
        }
        items = super().fetch(category,
                              filter_classified=filter_classified,
                              **kwargs)

        return items

    def fetch_items(self, category, **kwargs):
        """Fetch the messages.

        :param category: the category of items to fetch
        :param kwargs: backend arguments

        :returns: a generator of items
        """
        from_date = kwargs['from_date']
        to_date = kwargs['to_date']

        logger.info(f"Fetching messages on '{self.uri}' from {from_date} to {to_date}")

        group = self.client.group(self.group_tag)

        if not group:
            raise BackendError(cause=f'Group {self.uri} not found.')

        mailbox_id = group['archiveMailboxId']
        items = self.client.messages(mailbox_id, from_date, to_date)

        return items

    @classmethod
    def has_archiving(cls):
        """Returns whether it supports archiving items on the fetch process.

        :returns: this backend does not support items archive
        """
        return False

    @classmethod
    def has_resuming(cls):
        """Returns whether it supports to resume the fetch process.

        :returns: this backend supports items resuming
        """
        return True

    @staticmethod
    def metadata_id(item):
        """Extracts the identifier from a message item."""

        return str(item[Topicbox.MESSAGE_ID_FIELD][0])

    @staticmethod
    def metadata_updated_on(item):
        """Extracts the update time from a message item.

        The timestamp is extracted from 'Date' field.
        This date is converted to UNIX timestamp format.

        :param item: item generated by the backend

        :returns: a UNIX timestamp
        """
        ts = item[Topicbox.DATE_FIELD]
        ts = str_to_datetime(ts)

        return ts.timestamp()

    @staticmethod
    def metadata_category(item):
        """Extracts the category from a message item.

        This backend only generates one type of item which is
        'message'.
        """
        return CATEGORY_MESSAGE

    def _init_client(self, from_archive=False):
        """Init client"""

        base_uri = self.uri.split('/groups')[0]
        return TopicboxClient(base_uri=base_uri, account_id=self.account_id,
                              max_items=self.max_items)


class TopicboxClient(HttpClient):
    """Client for retrieving emails from Topicbox

    :param base_uri: Base URL of the Topicbox web application
    :param account_id: account id for Topicbox's JMAP
    :param max_items: max number of messages items per query
    """
    def __init__(self, base_uri, account_id, max_items=MAX_MESSAGES_PER_PAGE):
        self.account_id = account_id
        self.max_items = max_items
        self.jmap_uri = urijoin(base_uri, 'jmap')

        super().__init__(base_url=self.jmap_uri)

    def group(self, identifier):
        """Get group info associated to an identifier"""

        request = {
            "using": JMAP_CLIENT_CAPABILITIES,
            "methodCalls": [[
                "Group/get", {
                    "accountId": self.account_id,
                    "ids": None
                }, "0"
            ]]
        }
        res = self.fetch(self.jmap_uri, payload=json.dumps(request), method=HttpClient.POST)
        data = res.json()

        for group in data['methodResponses'][0][1]['list']:
            if group['identifier'] == identifier:
                return group

        return None

    def messages(self, mailbox_id, from_date=DEFAULT_DATETIME, to_date=DEFAULT_LAST_DATETIME):
        """Fetch Topicbox mailbox messages

        :param mailbox_id: archive Mailbox id for a group
        :param from_date: obtain messages created since this date
        :param to_date: obtain messages until a specific date
        """
        position = 0

        from_date_iso = from_date.strftime("%Y-%m-%dT%H:%M:%SZ")
        to_date_iso = to_date.strftime("%Y-%m-%dT%H:%M:%SZ")

        request = {
            "using": JMAP_CLIENT_CAPABILITIES,
            "methodCalls": [
                [
                    "Email/query", {
                        "accountId": self.account_id,
                        "calculateTotal": True,
                        "limit": self.max_items,
                        "position": position,
                        "collapseThreads": False,
                        "sort": [{"property": "receivedAt", "isAscending": True}],
                        "filter": {
                            "operator": "AND",
                            "conditions": [
                                {"inMailbox": mailbox_id},
                                {"after": from_date_iso},
                                {"before": to_date_iso}
                            ]
                        }
                    }, "0"
                ],
                [
                    "Email/get", {
                        "accountId": self.account_id,
                        "#ids": {"name": "Email/query", "path": "/ids", "resultOf": "0"},
                        "fetchHTMLBodyValues": True,
                        "properties": None
                    }, "1"
                ]
            ]
        }

        while True:
            request['methodCalls'][0][1]['position'] = position
            res = self.fetch(self.jmap_uri, payload=json.dumps(request), method=HttpClient.POST)
            data = res.json()

            messages = data['methodResponses'][1][1]['list']
            for message in messages:
                yield message

            total = data['methodResponses'][0][1]['total']
            position += len(messages)
            if position >= total:
                break


class TopicboxCommand(BackendCommand):
    """Class to run Topicbox backend from the command line."""

    BACKEND = Topicbox

    @classmethod
    def setup_cmd_parser(cls):
        """Returns the Topicbox argument parser."""

        parser = BackendCommandArgumentParser(cls.BACKEND,
                                              from_date=True,
                                              to_date=True)
        # Required arguments
        group = parser.parser.add_argument_group('Topicbox arguments')
        group.add_argument('uri',
                           help="URL of the Topicbox group")
        group.add_argument('--account-id', dest='account_id',
                           type=str, help="Account ID for Topicbox")
        group.add_argument('--max-items', dest='max_items',
                           default=MAX_MESSAGES_PER_PAGE, type=int,
                           help="Max number of messages per query.")

        return parser



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Jose Javier Merchante <jjmerchante@bitergia.com>
#

import sys
import unittest


if __name__ == '__main__':
    test_suite = unittest.TestLoader().discover('.', pattern='test_*.py')
    result = unittest.TextTestRunner(buffer=True).run(test_suite)
    sys.exit(not result.wasSuccessful())
